#!/usr/bin/env python2.5
# -*- coding: utf-8 -*-

"""
Thingspeak Plugin
plugin.py
Author: DaveL17
Credits:    Chris - http://www.australianrobotics.com.au/news/how-to-
                    talk-to-thingspeak-with-python-a-memory-cpu-monitor
            Karl (kw123) - device state restriction methods
            Update Checker by: berkinet (with additional features by
                    Travis Cook)

This script takes device and variable values and uploads them to
Thingspeak. In order to use it, you will need to have a Thingspeak
account, and have configured Thingspeak channels. Each Thingspeak
channel has a several API keys--this plugin uses an write key, which is
unique to each channel. To use the plugin, create individual Indigo
devices that represent a single Thingspeak channel. Under the device
configuration menu, enter the appropriate Thingspeak API key, and
assign Indigo devices or variables to each "Thing." Each plugin device
can hold up to eight variables (a Thingspeak limit, not the plugin's.
Currently, only numeric and binary variables will chart on Thingspeak,
and the plugin tries to adjust variable values as needed to make them
compatible (i.e., converting a string to a float.
"""

import datetime
import os.path
import simplejson
import subprocess
import time
import urllib  # Use to encode the URL prarameters.
import indigoPluginUpdateChecker

kDefaultPluginPrefs = {
    'configMenuTimeoutInterval': 15,          # How long to wait on a server timeout.
    'configMenuUploadInterval' : 900,         # How long to wait before refreshing devices.
    'elevation'                : 0,           # Elevation of data source.
    'latitude'                 : 0,           # Latitude of data source.
    'longitude'                : 0,           # Longitude of data source.
    'showDebugInfo'            : False,       # Verbose debug logging?
    'showDebugLevel'           : 1,           # Low, Medium or High debug output.
    'updaterEmail'             : "",          # Email to notify of plugin updates.
    'updaterEmailsEnabled'     : False,       # Notification of plugin updates wanted.
    'logFileDate'              : "1970-01-01",  # Log file creation date.
    }


class Plugin(indigo.PluginBase):
    def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
        indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
        self.debugLog(u"Thingspeak initialization called.")

        self.debug          = self.pluginPrefs.get('showDebugInfo', False)
        self.debugLevel     = self.pluginPrefs.get('showDebugLevel', 1)
        self.elevation      = self.pluginPrefs.get('elevation', 0)
        self.latitude       = self.pluginPrefs.get('latitude', 0)
        self.logFile        = '/Library/Application Support/Perceptive Automation/Indigo 6/Logs/Thingspeak Log.txt'
        self.logFileDate    = self.pluginPrefs.get('logFileDate', "1970-01-01")
        self.longitude      = self.pluginPrefs.get('longitude', 0)
        self.pluginName     = "com.fogbert.indigoplugin.thingspeak"
        self.uploadInterval = int(self.pluginPrefs.get('configMenuUploadInterval', 900))
        updater_url         = "https://dl.dropboxusercontent.com/u/2796881/thingspeak_version.html"
        self.updater        = indigoPluginUpdateChecker.updateChecker(self, updater_url)

        # Convert old debugLevel scale to new scale.
        # =============================================================
        if not 0 < self.pluginPrefs['showDebugLevel'] <= 3:
            if self.pluginPrefs['showDebugLevel'] == "High":
                self.pluginPrefs['showDebugLevel'] = 3
            elif self.pluginPrefs['showDebugLevel'] == "Medium":
                self.pluginPrefs['showDebugLevel'] = 2
            else:
                self.pluginPrefs['showDebugLevel'] = 1

        self.debugLevel = self.pluginPrefs.get('showDebugLevel', 1)

    def __del__(self):
        indigo.PluginBase.__del__(self)

    def startup(self):
        self.debugLog(u"Thingspeak startup() method called.")

        if self.debug and self.debugLevel >= 3:
            self.debugLog(u"Warning! Debug set to high. Sensitive information will be sent to the Indigo log.")

        # See if there is an update and whether the user wants to be notified.
        self.updater.checkVersionPoll()

    def shutdown(self):
        # Do any cleanup necessary before exiting
        self.debugLog(u"Thingspeak shutdown() method called.")
        pass

    def deviceStartComm(self, dev):
        self.debugLog(u"deviceStartComm() method called. Starting Thingspeak device: %s" % dev.name)
        dev.stateListOrDisplayStateIdChanged()
        dev.updateStateOnServer('thingState', value=False, uiValue=u"Enabled")
        dev.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)

    def deviceStopComm(self, dev):
        self.debugLog(u"deviceStopComm() method called. Stopping Thingspeak device: %s" % dev.name)
        dev.updateStateOnServer('thingState', value=False, uiValue=u"Disabled")

    def toggleDebugEnabled(self):
        # Toggle debug on/off.
        self.debugLog(u"toggleDebugEnabled() method called.")
        if not self.debug:
            self.debug = True
            self.pluginPrefs['showDebugInfo'] = True
            indigo.server.log(u"Debugging on.")
            self.debugLog(u"Debug level: %s" % self.debugLevel)
            if self.debugLevel >= 3:
                self.debugLog(u"Warning! Debug set to high. Sensitive information will be sent to the Indigo log.")

        else:
            self.debug = False
            self.pluginPrefs['showDebugInfo'] = False
            indigo.server.log(u"Debugging off.")

    def validatePrefsConfigUi(self, valuesDict):
        self.debugLog(u"validatePrefsConfigUi() method called.")

        error_msg_dict = indigo.Dict()

        try:
            # Test elevation. Must be an integer (can not be a float. Can be negative.
            try:
                int(valuesDict['elevation'])
            except:
                error_msg_dict['elevation'] = (
                    u"Thingspeak requires elevation to be expressed as "
                    "a whole number integer. It can be positive or negative.")
                return (False, valuesDict, error_msg_dict)

            if "." in valuesDict['elevation']:
                error_msg_dict['elevation'] = (
                    u"Thingspeak requires elevation to be expressed as "
                    "a whole number integer. It can be positive or negative.")
                return (False, valuesDict, error_msg_dict)

            # Test latitude and longitude. Must be integers or floats. Can be negative.
            try:
                float(valuesDict['latitude'])
            except:
                error_msg_dict['latitude'] = (
                    u"Thingspeak requires latitude to be expressed as "
                    "a whole number or decimal. It can be positive or negative.")
                return (False, valuesDict, error_msg_dict)

            try:
                float(valuesDict['longitude'])
            except:
                error_msg_dict['longitude'] = (
                    u"Thingspeak requires longitude to be expressed as "
                    "a whole number or decimal. It can be positive or negative.")
                return (False, valuesDict, error_msg_dict)

            # Test plugin update notification settings.
            if valuesDict['updaterEmailsEnabled'] and not valuesDict['updaterEmail']:
                error_msg_dict['updaterEmail'] = (
                    u"If you want to be notified of updates, you must "
                    "supply an email address.")
                return (False, valuesDict, error_msg_dict)

            elif valuesDict['updaterEmailsEnabled'] and "@" not in valuesDict['updaterEmail']:
                error_msg_dict['updaterEmail'] = (
                    u"Valid email addresses have at leat one @ symbol "
                    "in them (foo@bar.com).")
                return (False, valuesDict, error_msg_dict)

        except Exception, e:

            f = open(self.logFile, 'a')
            f.write(u"%s - validatePrefsConfigUi error: %s\n" % (datetime.datetime.time(datetime.datetime.now()), e))
            f.close

            self.debugLog(u"Exception in validatePrefsConfigUi tests: %s" % e)
            pass

        return (True, valuesDict)

    def closedPrefsConfigUi(self, valuesDict, userCancelled):
        self.debugLog(u"closedPrefsConfigUi() method called.")

        if userCancelled:
            self.debugLog(u"User prefs dialog cancelled.")

        if not userCancelled:
            self.debug = valuesDict.get('showDebugInfo', False)
            self.debugLog(u"User prefs saved.")
            if self.pluginPrefs['showDebugLevel'] >= 3:
                self.debugLog(str(valuesDict))

            if self.debug:
                indigo.server.log(u"Debug logging is on.")
                if self.debugLevel >= 3:
                    self.debugLog(u"Warning! Debug set to high. Sensitive information will be sent to the Indigo log.")

        # Update update select globals if plugin prefs have changed.
        self.debugLevel     = self.pluginPrefs.get('showDebugLevel', 1)
        self.elevation      = self.pluginPrefs.get('elevation', 0)
        self.latitude       = self.pluginPrefs.get('latitude', 0)
        self.longitude      = self.pluginPrefs.get('longitude', 0)
        self.uploadInterval = int(self.pluginPrefs.get('configMenuUploadInterval', 900))

    def checkDebugLogFile(self):
        '''
        The checkDebugLogFile() method manages the Thingspeak Plugin
        logging facility. It ensures the log file always exists and
        that it never contains more than one day of logs.
        '''
        self.debugLog(u"checkDebugLogFile() method called.")
        log     = datetime.datetime.strptime(self.pluginPrefs.get('logFileDate', "1970-01-01"), "%Y-%m-%d")
        logDate = datetime.datetime.date(log)
        now     = datetime.datetime.now()
        today   = datetime.datetime.date(now)

        # Create the log file if it doesn't exist or if it's a new day,
        # delete the existing log file and create a new one.
        if not os.path.exists(self.logFile) or today > logDate:
            f = open(self.logFile, 'w')
            f.write((u'*' * 72) + "\n")
            f.write('***%s- Thingspeak Log - %s - %s***\n' % ((u' ' * 16), today, (u' ' * 18)))
            f.write((u'*' * 72) + "\n")
            f.close()

            # Update plugin log date tracker with today's date.
            self.pluginPrefs['logFileDate'] = str(today)

    def checkVersionNow(self):
        self.debugLog(u"checkVersionNow() method called.")
        self.updater.checkVersionNow()

    def listGenerator(self, filter="", valuesDict=None, typeId="", targetId=0):
        '''This method collects IDs and names for all Indigo devices and
        variables. It creates a dictionary of the form
        ((dev.id, dev.name), (var.id, var.name)).'''
        self.debugLog(u"listGenerator() method called.")

        dev_list = [(dev.id, "(D) %s" % dev.name) for dev in indigo.devices]
        var_list = [(var.id, "(V) %s" % var.name) for var in indigo.variables]
        List     = dev_list + var_list
        List.append(('None', 'None'))
        if self.debugLevel >= 3:
            self.debugLog(u"Generated list of devices and variables:")
            self.debugLog(unicode(List))
        return List

    def deviceStateGenerator1(self, filter="", valuesDict=None, typeId="", targetId=0):
        '''
        The following eight methods produce lists of states that are
        associated with user-selected devices when configuring
        Thingspeak reporting devices. Each list includes only states
        for the selected device.
        '''
        self.debugLog(u"deviceStateGenerator1 method called.")

        state_list_1 = list()

        # If there are no Thingspeak devices created yet.
        if not valuesDict:
            return state_list_1

        if valuesDict and "thing1" in valuesDict:

            # If an item has been selected.
            if valuesDict['thing1'] != "" and valuesDict['thing1'] != "None":
                device_string = int(valuesDict['thing1'])

                try:
                    # If it's a device, grab the selected state.
                    for devID in indigo.devices.iterkeys():

                        if device_string == devID:

                            # If there's no device specified, the state
                            # is NoneType.
                            for state in indigo.devices[device_string].states:

#                                 if ".ui" in state or "All" in state or "zone" in state:
                                if any(item in state for item in (".ui", "All", "zone")):
                                    pass
                                else:
                                    state_list_1.append(state)

                    # If it's not a device, it's a variable. Grab it's value.
                    for varID in indigo.variables.iterkeys():

                        if device_string == varID:
                            var = indigo.variables[device_string].name
                            state_list_1.append(var + u" value")

                # If it's somehow not a device or a variable, skip it.
                except:
                    indigo.server.log(u"Element not of device type or variable type. Skipping.")
                    pass

                # Append the list with a "None" choice.
                state_list_1.append("None")
                return state_list_1

            # If an item has not been selected, return and empty list.
            else:
                return state_list_1

    def deviceStateGenerator2(self, filter="", valuesDict=None, typeId="", targetId=0):
        self.debugLog(u"deviceStateGenerator2 method called.")
        state_list_2 = list()

        if not valuesDict:
            return state_list_2

        if valuesDict and "thing2" in valuesDict:
            if valuesDict['thing2'] != "" and valuesDict['thing2'] != "None":
                device_string = int(valuesDict['thing2'])
                try:
                    for devID in indigo.devices.iterkeys():
                        if device_string == devID:
                            for state in indigo.devices[device_string].states:
#                                 if ".ui" in state or "All" in state or "zone" in state:
                                if any(item in state for item in (".ui", "All", "zone")):
                                    pass
                                else:
                                    state_list_2.append(state)

                    for varID in indigo.variables.iterkeys():
                        if device_string == varID:
                            var = indigo.variables[device_string].name
                            state_list_2.append(var + u" value")
                except:
                    pass

                state_list_2.append("None")
                return state_list_2

            else:
                return state_list_2

    def deviceStateGenerator3(self, filter="", valuesDict=None, typeId="", targetId=0):
        self.debugLog(u"deviceStateGenerator3 method called.")
        state_list_3 = list()

        if not valuesDict:
            return state_list_3

        if valuesDict and "thing3" in valuesDict:
            if valuesDict['thing3'] != "" and valuesDict['thing3'] != "None":
                device_string = int(valuesDict['thing3'])
                try:
                    for devID in indigo.devices.iterkeys():
                        if device_string == devID:
                            for state in indigo.devices[device_string].states:
#                                 if ".ui" in state or "All" in state or "zone" in state:
                                if any(item in state for item in (".ui", "All", "zone")):
                                    pass
                                else:
                                    state_list_3.append(state)

                    for varID in indigo.variables.iterkeys():
                        if device_string == varID:
                            var = indigo.variables[device_string].name
                            state_list_3.append(var + u" value")
                except:
                    pass

                state_list_3.append("None")
                return state_list_3

            else:
                return state_list_3

    def deviceStateGenerator4(self, filter="", valuesDict=None, typeId="", targetId=0):
        self.debugLog(u"deviceStateGenerator4 method called.")
        state_list_4 = list()

        if not valuesDict:
            return state_list_4

        if valuesDict and "thing4" in valuesDict:
            if valuesDict['thing4'] != "" and valuesDict['thing4'] != "None":
                device_string = int(valuesDict['thing4'])
                try:
                    for devID in indigo.devices.iterkeys():
                        if device_string == devID:
                            for state in indigo.devices[device_string].states:
#                                 if ".ui" in state or "All" in state or "zone" in state:
                                if any(item in state for item in (".ui", "All", "zone")):
                                  pass
                                else:
                                    state_list_4.append(state)

                    for varID in indigo.variables.iterkeys():
                        if device_string == varID:
                            var = indigo.variables[device_string].name
                            state_list_4.append(var + u" value")
                except:
                    pass

                state_list_4.append("None")
                return state_list_4

            else:
                return state_list_4

    def deviceStateGenerator5(self, filter="", valuesDict=None, typeId="", targetId=0):
        self.debugLog(u"deviceStateGenerator5 method called.")
        state_list_5 = list()

        if not valuesDict:
            return state_list_5

        if valuesDict and "thing5" in valuesDict:

            if valuesDict['thing5'] != "" and valuesDict['thing5'] != "None":
                device_string = int(valuesDict['thing5'])
                try:
                    for devID in indigo.devices.iterkeys():
                        if device_string == devID:
                            for state in indigo.devices[device_string].states:
#                                 if ".ui" in state or "All" in state or "zone" in state:
                                if any(item in state for item in (".ui", "All", "zone")):
                                    pass
                                else:
                                    state_list_5.append(state)

                    for varID in indigo.variables.iterkeys():
                        if device_string == varID:
                            var = indigo.variables[device_string].name
                            state_list_5.append(var + u" value")
                except:
                    pass

                state_list_5.append("None")
                return state_list_5

            else:
                return state_list_5

    def deviceStateGenerator6(self, filter="", valuesDict=None, typeId="", targetId=0):
        self.debugLog(u"deviceStateGenerator6 method called.")
        state_list_6 = list()

        if not valuesDict:
            return state_list_6

        if valuesDict and "thing6" in valuesDict:
            if valuesDict['thing6'] != "" and valuesDict['thing6'] != "None":
                device_string = int(valuesDict['thing6'])
                try:
                    for devID in indigo.devices.iterkeys():
                        if device_string == devID:
                            for state in indigo.devices[device_string].states:
#                                 if ".ui" in state or "All" in state or "zone" in state:
                                if any(item in state for item in (".ui", "All", "zone")):
                                    pass
                                else:
                                    state_list_6.append(state)

                    for varID in indigo.variables.iterkeys():
                        if device_string == varID:
                            var = indigo.variables[device_string].name
                            state_list_6.append(var + u" value")
                except:
                    pass

                state_list_6.append("None")
                return state_list_6

            else:
                return state_list_6

    def deviceStateGenerator7(self, filter="", valuesDict=None, typeId="", targetId=0):
        self.debugLog(u"deviceStateGenerator7 method called.")
        state_list_7 = list()

        if not valuesDict:
            return state_list_7

        if valuesDict and "thing7" in valuesDict:
            if valuesDict['thing7'] != "" and valuesDict['thing7'] != "None":
                device_string = int(valuesDict['thing7'])
                try:
                    for devID in indigo.devices.iterkeys():
                        if device_string == devID:
                            for state in indigo.devices[device_string].states:
#                                 if ".ui" in state or "All" in state or "zone" in state:
                                if any(item in state for item in (".ui", "All", "zone")):
                                    pass
                                else:
                                    state_list_7.append(state)

                    for varID in indigo.variables.iterkeys():
                        if device_string == varID:
                            var = indigo.variables[device_string].name
                            state_list_7.append(var + u" value")
                except:
                    pass

                state_list_7.append("None")
                return state_list_7

            else:
                return state_list_7

    def deviceStateGenerator8(self, filter="", valuesDict=None, typeId="", targetId=0):
        self.debugLog(u"deviceStateGenerator8 method called.")
        state_list_8 = list()

        if not valuesDict:
            return state_list_8

        if valuesDict and "thing8" in valuesDict:
            if valuesDict['thing8'] != "" and valuesDict['thing8'] != "None":
                device_string = int(valuesDict['thing8'])
                try:
                    for devID in indigo.devices.iterkeys():
                        if device_string == devID:
                            for state in indigo.devices[device_string].states:
#                                 if ".ui" in state or "All" in state or "zone" in state:
                                if any(item in state for item in (".ui", "All", "zone")):
                                    pass
                                else:
                                    state_list_8.append(state)

                    for varID in indigo.variables.iterkeys():
                        if device_string == varID:
                            var = indigo.variables[device_string].name
                            state_list_8.append(var + u" value")
                except:
                    pass

                state_list_8.append("None")
                return state_list_8

            else:
                return state_list_8

    def encodeValueDicts(self):

        # Encode dicts of data for [type:Thingspeak] for upload.
        self.debugLog(u"encodeValueDicts() method called.")

        for dev in indigo.devices.itervalues("self"):

            # There are no Thingspeak devices yet.
            if not self.pluginName:
                self.debugLog(u"No Thingspeak devices have been created yet.")
                self.sleep(self.uploadInterval-5)

            # A device has been created, but hasn't been saved yet.
            elif not dev.configured:
                indigo.server.log(
                    u"A device is being (or has been) created, but "
                    "it's not fully configured. Sleeping while you finish.")
                self.sleep(self.uploadInterval-5)

            # A Thingspeak device has no API key.
            elif not (
                    dev.globalProps[self.pluginName]['apiKey'] or
                    dev.globalProps[self.pluginName]['apiKey'] == ""):

                indigo.server.log(u"Device %s [%s] has no API Key. Skipping device." % (str(dev.name), str(dev.id)))

            elif not dev.enabled:
                indigo.server.log(u"%s device is disabled. Skipping." % dev.name)

            # Iterate over up to 8 values per device.
            elif dev.enabled:
                api_key    = dev.globalProps[self.pluginName]['apiKey']
                api_key    = self.fixApiKey(dev, api_key)
                thing_dict = {}

                for v in range(8):
                    v += 1
                    v_str           = str(v)
                    thing_str       = ('thing%s' % v_str)
                    thing_state_str = ('thing%sState' % v_str)

                    # Create the dict and add the API key to it.
                    thing_dict['key'] = api_key

                    # If there is a device created, but no value assigned.
                    if not (
                            dev.globalProps[self.pluginName][thing_str] or
                            dev.globalProps[self.pluginName][thing_str] == "None"):
                        var = "Null value"

                    else:
                        thing_1 = dev.globalProps[self.pluginName][thing_str]
                        state_1 = dev.globalProps[self.pluginName][thing_state_str]

                        self.debugLog(u"" * 22)
                        self.debugLog(unicode(u"ID: %s" % thing_1))
                        self.debugLog(unicode(u"Item: %s" % state_1))

                        # If it's a device state, do this:
                        if int(thing_1) in indigo.devices:

                            try:
                                var = indigo.devices[int(thing_1)].states[state_1]
                                var = self.onlyNumerics(var)
                                self.debugLog(u"Value: %s" % str(var))

                            except:
                                self.errorLog(
                                    u"%s - %s is non-numeric or has been removed. "
                                    "Will try to upload to Thingspeak, but it won't chart." %
                                    (dev.name, dev.globalProps[self.pluginName][thing_str]))

                            # Add device state value to dictionary.
                            thing_dict['field' + v_str] = var

                        # If it's a variable value, do this:
                        elif int(thing_1) in indigo.variables:
                            var = indigo.variables[int(thing_1)].value

                            try:
                                var = self.onlyNumerics(var)
                                self.debugLog(u"Value: %s" % str(var))

                            except:
                                self.errorLog(
                                    u"%s - %s is non-numeric or has been removed. "
                                    "Will try to upload to Thingspeak, but it won't chart." %
                                    (dev.name, dev.globalProps[self.pluginName][thing_str]))

                            # Add variable value to dictionary.
                            thing_dict['field' + v_str] = var

                    thing_dict['elevation'] = self.elevation
                    thing_dict['latitude']  = self.latitude
                    thing_dict['longitude'] = self.longitude

                if self.debugLevel >= 2:
                    self.debugLog(unicode(thing_dict))

                # Open a connection and upload data to
                # Thingspeak
                try:

                    params = urllib.urlencode(thing_dict)
                    self.debugLog(u"Channel updating...")

                    # The plugin uploads variable values before
                    # moving on to the next one. Will continue
                    # until no more devices or the plugin
                    # throws an exception.
                    self.uploadToThingspeak(dev, params)

                except Exception, e:

                    f = open(self.logFile, 'a')
                    f.write(u"%s - Curl Return Code: %s\n %s \n" % (datetime.datetime.time(datetime.datetime.now()), e))
                    f.close

                    self.errorLog(str(e))

        return

    def fixApiKey(self, dev, api_key):
        '''
        This method evaluates user-provided Thingspeak api WRITE keys
        for form and length. Where possible (and appropriate), it
        corrects common errors such as length, and leading/trailing
        spaces. It then writes the repaired api_key back to the
        device dict.
        '''
        self.debugLog(u"fixApiKey() method called.")

        if api_key.startswith(" ") or api_key.endswith(" "):
            api_key = api_key.strip()
            self.errorLog(u"%s API key includes leading and/or trailing spaces. Repairing." % str(dev.name))

            # Overwrite with repaired api_key as needed.
            new_props           = dev.pluginProps
            new_props['apiKey'] = api_key
            dev.replacePluginPropsOnServer(new_props)

        # Thingspeak keys are always 16 digits in length. Plugin will
        # continue to run, but Thingspeak will reject/ignore (which?)
        # the submission.
        if len(api_key) != 16:
            self.errorLog(u"API key value %s for [%s] is not the proper length. Check key." % (api_key, dev.name))
        return api_key

    def onlyNumerics(self, seq):
        '''
        This method evaluates values intended for upload. It ensures
        that the values are numeric only (i.e., stripping °F) and
        converting binary strings to integers. It also converts
        Indigo string values to floats as necessary.
        '''
        self.debugLog(u"onlyNumerics() method called.")

        x = ""

        try:
            # Does it float? Yes? Then it must be a witch.
            x = float(seq)
            return x
        except:
            xx = ''.join([c for c in seq if c in '1234567890.'])
            if len(''.join([c for c in xx if c in '1234567890'])) != 0:
                return xx
            seq = str(seq).upper()

        if seq.lower() or seq.upper() in ("TRUE", "ON"):
            x = 1

        if seq.lower or seq.upper() in ("FALSE", "OFF"):
            x = 0

        return x

    def stopSleep(self, startSleep):
        '''
        The stopSleep() method accounts for changes to the user
        upload interval preference. The plugin checks every 2 seconds
        to see if the sleep interval should be updated.
        '''
        self.debugLog(u"stopSleep() method called.")
        try:
            # We subtract an additional 5 seconds to account for the 5
            # second sleep at the start of runConcurrentThread.
            totalSleep = float(self.pluginPrefs.get('configMenuUploadInterval', 900)) - 6
        except:
            totalSleep = iTimer
        if time.time() - startSleep > totalSleep:
            return True
        return False

    def updateThingspeakDataAction(self, valuesDict):
        '''
        The updateThingspeakDataAction () method invokes an instantaneous
        update of the Thingspeak data channels. If this is called before
        15 seconds have elapsed since the last update, Thingspeak will
        ignore it. Unsure if the 15 second limit starts over.
        '''
        self.debugLog(u"updateThingspeakDataAction() method called.")
        self.encodeValueDicts()
        return

    def updateThingspeakDataMenu(self):
        '''
        The updateThingspeakDataMenu() method invokes an instantaneous
        update of the Thingspeak data channels. If this is called before
        15 seconds have elapsed since the last update, Thingspeak will
        ignore it. Unsure if the 15 second limit starts over.
        '''
        self.debugLog(u"updateThingspeakDataMenu called.")
        self.encodeValueDicts()
        return

    def uploadToThingspeak(self, dev, params):
        '''
        This method performs the upload to Thingspeak, evaluates, and
        logs the result.
        '''
        self.debugLog(u"uploadToThingspeak() method called.")

        try:

            # Build upload URL.
            if dev.globalProps[self.pluginName]['devicePort']:
                TS_IP = dev.globalProps[self.pluginName]['deviceIP']
            else:
                TS_IP = "api.thingspeak.com"

            url = "https://%s/update.json?%s" % (TS_IP, params)
            if self.debugLevel >= 3:
                self.debugLog(u"Warning! Debug set to high. Upload URL containing your API key(s) written was written to debug output.")
                self.debugLog(url)
            else:
                self.debugLog(u"URL debug logging suppressed. Set debug level to high to write it to the log.")

            # Initiate curl call to Thingspeak servers.
            proc = subprocess.Popen(["curl", '-vs', url], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            (result, err) = proc.communicate()

            # Process the results.
            # Thingspeak will respond with a "0" if they something went wrong.
            if len(result) > 0 and result != "0":

                result = simplejson.loads(result, encoding="utf-8")

                for key, value in result.iteritems():
                    if not value:
                        result[key] = "0"

                dev.updateStateOnServer('channel_id', value=int(result['channel_id']), uiValue=unicode(result['channel_id']))
                dev.updateStateOnServer('elevation', value=int(result['elevation']), uiValue=unicode(result['elevation']))
                dev.updateStateOnServer('entry_id', value=int(result['entry_id']), uiValue=unicode(result['entry_id']))
                dev.updateStateOnServer('latitude', value=float(result['latitude']), uiValue=unicode(result['latitude']))
                dev.updateStateOnServer('longitude', value=float(result['longitude']), uiValue=unicode(result['longitude']))
                dev.updateStateOnServer('status', value=result['status'], uiValue=result['status'])
                dev.updateStateOnServer('thing1', value=result['field1'], uiValue=result['field1'])
                dev.updateStateOnServer('thing2', value=result['field2'], uiValue=result['field2'])
                dev.updateStateOnServer('thing3', value=result['field3'], uiValue=result['field3'])
                dev.updateStateOnServer('thing4', value=result['field4'], uiValue=result['field4'])
                dev.updateStateOnServer('thing5', value=result['field5'], uiValue=result['field5'])
                dev.updateStateOnServer('thing6', value=result['field6'], uiValue=result['field6'])
                dev.updateStateOnServer('thing7', value=result['field7'], uiValue=result['field7'])
                dev.updateStateOnServer('thing8', value=result['field8'], uiValue=result['field8'])
#                 dev.updateStateOnServer('twitter', value=result['twitter'], uiValue=result['twitter'])
#                 dev.updateStateOnServer('tweet', value=result['tweet'], uiValue=result['tweet'])

                # Convert UTC return to local time. There is an optional
                # timezone parameter that can be used in the form of:
                # time_zone="timezone=America%2FChicago&"
                # For now, we will convert to UTC locally.
                if result['created_at']:
                    t   = time.time()
                    utc = result['created_at']

                    # time_delta_to_utc formula thanks to Karl (kw123).
                    time_delta_to_utc = (int(time.mktime(datetime.datetime.utcfromtimestamp(t + 10).timetuple()) - t) / 100) * 100
                    utc_obj           = datetime.datetime.strptime(utc, '%Y-%m-%dT%H:%M:%SZ')
                    local_time        = str(utc_obj - datetime.timedelta(seconds = time_delta_to_utc))
                    dev.updateStateOnServer('created_at', value=local_time, uiValue=local_time)
                else:
                    dev.updateStateOnServer('created_at', value=u"Unknown", uiValue=u"Unknown")

                dev.updateStateOnServer('thingState', value=True, uiValue=" ")
                dev.updateStateImageOnServer(indigo.kStateImageSel.SensorOn)
                return

            # Something didn't go the way we wanted it to.
            if err:
                if proc.returncode == 6:

                    f = open(self.logFile, 'a')
                    f.write(u"%s - uploadToThingSpeak()\nCurl Return Code: %s\n%s \n" % (datetime.datetime.time(datetime.datetime.now()), proc.returncode, err))
                    f.close

                    self.errorLog(u"Error: Could not resolve host. Possible causes:")
                    self.errorLog(u"  The Thingspeak service is offline.")
                    self.errorLog(u"  Your Indigo server can not reach the Internet.")
                    self.errorLog(u"  Your plugin is misconfigured.")
                    self.debugLog(err)
                    dev.updateStateOnServer('thingState', value=False, uiValue="no comm")
#                     dev.updateStateImageOnServer(indigo.kStateImageSel.Error)
                    dev.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)

                elif err != "":

                    f = open(self.logFile, 'a')
                    f.write(u"%s - Curl Return Code: %s\n %s \n" % (datetime.datetime.time(datetime.datetime.now()), proc.returncode, err))
                    f.close

                    self.debugLog(u"\n" + err)
#                     dev.updateStateImageOnServer(indigo.kStateImageSel.Error)
                    dev.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)
                    dev.updateStateOnServer('thingState', value=False, uiValue="error")
                return

        # Something didn't go the way we wanted it to that we didn't anticipate.
        except Exception, e:

            f = open(self.logFile, 'a')
            f.write(u"%s - Misc Exception: %s\n" % (datetime.datetime.time(datetime.datetime.now()), e))
            f.close

            self.errorLog(u"Unable to upload Thingspeak data. Reason: Exception - %s" % unicode(e))
#             dev.updateStateImageOnServer(indigo.kStateImageSel.Error)
            dev.updateStateImageOnServer(indigo.kStateImageSel.SensorOff)
            dev.updateStateOnServer('thingState', value=False, uiValue="error")
            pass

        return

    def runConcurrentThread(self):
        self.debugLog(u"runConcurrentThread initiated.")

        try:
            while True:
                self.sleep(5)

                self.updater.checkVersionPoll()
                self.checkDebugLogFile()
                self.encodeValueDicts()

                startSleep = time.time()
                while True:
                    if self.stopSleep(startSleep):
                        break
                    self.sleep(2)

        except self.StopThread:
            # Do any cleanup necessary before stopping
            self.debugLog(u"Thingspeak stop thread called.")
            pass
